/* 
 * keyboard.c --
 *
 *	This file contains code that implements xop's keyboard subcommand.
 */

/*
 * Copyright (C) 1998-2001, Neil McKay
 */

#include "tk.h"

/* Include all the X11 stuff we need. */

#include <X11/Xutil.h>
#ifdef XF86MISC
#include <X11/Xproto.h>
#include <X11/extensions/xf86misc.h>
#include <X11/extensions/xf86mscstr.h>
#endif

#include "xopInt.h"


/*
 *--------------------------------------------------------------
 *
 * XopKeyboardCommand --
 *
 *	Implements the "keyboard" Xop subcommand. See the Xop
 *	man page for a description.
 *
 * Results:
 *	A standard Tcl result.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

int
XopKeyboardCommand(interp, argc, argv, opts)
    Tcl_Interp *interp;	/* Result goes here	*/
    int argc;		/* Argument count	*/
    char **argv;	/* Argument vector	*/
    XopOptions *opts;	/* Options		*/
{
    int minKeycode, maxKeycode;
    char junkStr[32];
    Tcl_DString ds;
    int i, j, idx, keysymIdx;
    int bool;
    unsigned int mask;
    KeySym keysym;
    int minMapKeycode, maxMapKeycode, maxKeysymsPerKeycode;
    int keysymsPerKeycode;
    KeySym *keySymList;
    int nItems;
    char **itemList;
    int keycode, keyCode;
    int keycodeCount;
    int nKeysyms;
    char **keysymNameList;
    int offset;
    KeySym *currentSymList;
    int currentKeycodeCount;
    int currentKeysymsPerKeycode;
    int newMapSize, newMapNumEntries;
    KeySym newKeysym, *newKeyMap;
    int keycodeOffset;
    XKeyboardState kbState;
    XKeyboardControl kbControl;
    int clickVolume;
    unsigned long ledMask;
    int ledNo;
    int ledOn;
    char keyStateReturn[32];
#ifdef XF86MISC
    int major, minor;
    XF86MiscKbdSettings kbdSettings;
#endif

    if (argc < 3) {
	Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
		" <subcommand>\nValid subcommands are:\n",
		"	autorepeat click coderange leds mapping state",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "coderange") == 0) {
	if (argc != 3) {
	    Tcl_AppendResult(interp,
			"Extraneous arguments follow \"coderange\"",
			(char *) NULL);
	    return TCL_ERROR;
	}

	XDisplayKeycodes(opts->display, &minKeycode, &maxKeycode);

	sprintf(junkStr, "%1d", minKeycode);
	Tcl_AppendElement(interp, junkStr);
	sprintf(junkStr, "%1d", maxKeycode);
	Tcl_AppendElement(interp, junkStr);
	return TCL_OK;
    }

    if (strcmp(argv[2], "autorepeat") == 0) {
	if (argc == 3) {
	    /* Return the global autorepeat setting */

	    XGetKeyboardControl(opts->display, &kbState);
	    if (kbState.global_auto_repeat == AutoRepeatModeOn) {
		Tcl_SetResult(interp, "on", TCL_STATIC);
	    } else {
		Tcl_SetResult(interp, "off", TCL_STATIC);
	    }

	    return TCL_OK;
	}

	if (strcmp(argv[3], "on") == 0
			|| strcmp(argv[3], "off") == 0
			|| strcmp(argv[3], "default") == 0) {
	    if (argc != 4) {
		Tcl_AppendResult(interp,
			"Extraneous arguments follow \"", argv[3],
			"\"", (char *) NULL);
		return TCL_ERROR;
	    }
	    if (strcmp(argv[3], "on") == 0) {
		kbControl.auto_repeat_mode = AutoRepeatModeOn;
	    } else if (strcmp(argv[3], "off") == 0) {
		kbControl.auto_repeat_mode = AutoRepeatModeOff;
	    } else {
		kbControl.auto_repeat_mode = AutoRepeatModeDefault;
	    }

	    XChangeKeyboardControl(opts->display, KBAutoRepeatMode, &kbControl);
	    return TCL_OK;
	}

	if (strcmp(argv[3], "map") == 0) {
	    if (argc > 5) {
		Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
			argv[1], " ", argv[2], " ", argv[3],
			" ?autoRepeatMap?", (char *) NULL);
		return TCL_ERROR;
	    }

	    /* Get the valid keycode range for this display. */

	    XDisplayKeycodes(opts->display, &minKeycode, &maxKeycode);

	    if (argc == 4) {
		/* Get the autorepeat map. */

		XGetKeyboardControl(opts->display, &kbState);

		/* Write the autorepeat map into interp. */

		for(keycode = minKeycode; keycode <= maxKeycode; keycode++) {
		    idx = keycode%8;
		    mask = 1 << (keycode & 0x07);
		    sprintf(junkStr, "%1d", keycode);
		    Tcl_AppendElement(interp, junkStr);
		    if (kbState.auto_repeats[idx] & mask) {
			Tcl_AppendElement(interp, "on");
		    } else {
			Tcl_AppendElement(interp, "off");
		    }
		}

		return TCL_OK;
	    }

	    /* We're setting the autorepeat map. */

	    if (Tcl_SplitList(interp, argv[4], &nItems, &itemList) != TCL_OK) {
		return TCL_ERROR;
	    }

	    if (nItems%2 != 0) {
		ckfree((VOID *) itemList);
		Tcl_SetResult(interp,
			"autorepeat map must have an even number of elements",
			TCL_STATIC);
		return TCL_ERROR;
	    }

	    for (idx = 0; idx < nItems; idx += 2) {
		if (Tcl_GetInt(interp, itemList[idx], &keycode) != TCL_OK) {
		    ckfree((VOID *) itemList);
		    return TCL_ERROR;
		}
		if (keycode < minKeycode || keycode > maxKeycode) {
		    ckfree((VOID *) itemList);
		    Tcl_AppendResult(interp,
				"keycode ", itemList[idx],
				" out of range", (char *) 0);
		    return TCL_ERROR;
		}

		if (strcmp(itemList[idx+1], "default") == 0) {
		    kbControl.auto_repeat_mode = AutoRepeatModeDefault;
		} else {
		    if (Tcl_GetBoolean(interp,
				itemList[idx+1], &bool) != TCL_OK) {
			ckfree((VOID *) itemList);
			return TCL_ERROR;
		    }
		    if (bool) {
			kbControl.auto_repeat_mode = AutoRepeatModeOn;
		    } else {
			kbControl.auto_repeat_mode = AutoRepeatModeOn;
		    }
		}

		kbControl.key = keycode;
		XChangeKeyboardControl(opts->display,
			KBKey|KBAutoRepeatMode, &kbControl);
	    }

	    return TCL_OK;
	}

#ifdef XF86MISC
	if (strcmp(argv[3], "delay") == 0) {
	    if (argc > 5) {
		Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
			argv[1], " ", argv[2], " ", argv[3],
			" ?delay?", (char *) NULL);
		return TCL_ERROR;
	    }

	    /* Make sure the server has this extension. */

	    if (!XF86MiscQueryVersion(opts->display, &major, &minor)) {
		Tcl_AppendResult(interp,
			"Display does not support XF86MISC extension",
			(char *) NULL);
		return TCL_ERROR;
	    }

	    /* Get the current settings. */

	    XF86MiscGetKbdSettings(opts->display, &kbdSettings);

	    /* See if we're supposed to return the delay. */

	    if (argc == 4) {
		sprintf(junkStr, "%1d", kbdSettings.delay);
		Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
		return TCL_OK;
	    }

	    /* Set the delay. */

	    if (Tcl_GetInt(interp, argv[4], &i) != TCL_OK) {
		return TCL_ERROR;
	    }

	    kbdSettings.delay = i;
	    XF86MiscSetKbdSettings(opts->display, &kbdSettings);
	    return TCL_OK;
	}

	if (strcmp(argv[3], "rate") == 0) {
	    if (argc > 5) {
		Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
			argv[1], " ", argv[2], " ", argv[3],
			" ?rate?", (char *) NULL);
		return TCL_ERROR;
	    }

	    /* Make sure the server has this extension. */

	    if (!XF86MiscQueryVersion(opts->display, &major, &minor)) {
		Tcl_AppendResult(interp,
			"Display does not support XF86MISC extension",
			(char *) NULL);
		return TCL_ERROR;
	    }

	    /* Get the current settings. */

	    XF86MiscGetKbdSettings(opts->display, &kbdSettings);

	    /* See if we're supposed to return the repeat rate. */

	    if (argc == 4) {
		sprintf(junkStr, "%1d", kbdSettings.rate);
		Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
		return TCL_OK;
	    }

	    /* Set the repeat rate. */

	    if (Tcl_GetInt(interp, argv[4], &i) != TCL_OK) {
		return TCL_ERROR;
	    }

	    kbdSettings.rate = i;
	    XF86MiscSetKbdSettings(opts->display, &kbdSettings);
	    return TCL_OK;
	}
#endif

	Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
		argv[1], " ", argv[2], " ?on|off|default|delay|map|rate? ...",
		(char *) NULL);
	return TCL_ERROR;
    }

    if (strcmp(argv[2], "click") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
			argv[1], " ", argv[2], " ?volume?", (char *) NULL);
	    return TCL_ERROR;
	}

	if (argc == 3) {
	    XGetKeyboardControl(opts->display, &kbState);
	    sprintf(junkStr, "%1d", kbState.key_click_percent);
	    Tcl_SetResult(interp, junkStr, TCL_VOLATILE);
	    return TCL_OK;
	}

	if (Tcl_GetInt(interp, argv[3], &clickVolume) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (clickVolume < -1) {
	    clickVolume = -1;
	} else if (clickVolume > 100) {
	    clickVolume = 100;
	}

	kbControl.key_click_percent = clickVolume;
	XChangeKeyboardControl(opts->display, KBKeyClickPercent, &kbControl);

	return TCL_OK;
    }

    if (strcmp(argv[2], "leds") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?ledMap?", (char *) NULL);
	    return TCL_ERROR;
	}

	if (argc == 3) {
	    /* Get the state of the keyboard LEDs. */

	    XGetKeyboardControl(opts->display, &kbState);
	    ledMask = 1;
	    for (i = 1; i <= 32; i++, ledMask <<= 1) {
		sprintf(junkStr, "%1d", i);
		Tcl_AppendElement(interp, junkStr);
		if (kbState.led_mask & ledMask) {
		    Tcl_AppendElement(interp, "on");
		} else {
		    Tcl_AppendElement(interp, "off");
		}
	    }

	    return TCL_OK;
	}

	/* Set the state of the keyboard leds. */

	if (Tcl_SplitList(interp, argv[3], &nItems, &itemList) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (nItems%2 != 0) {
	    Tcl_SetResult(interp,
		"LED map must have an even number of elements", TCL_STATIC);
	}

	for (i = 0; i < nItems; i += 2) {
	    if (Tcl_GetInt(interp, itemList[i], &ledNo) != TCL_OK) {
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    if (ledNo < 1 || ledNo > 32) {
		ckfree((VOID *) itemList);
		Tcl_SetResult(interp,
			"LED number must be between 1 and 32", TCL_STATIC);
		return TCL_ERROR;
	    }

	    kbControl.led = ledNo;

	    if (Tcl_GetBoolean(interp, itemList[i+1], &ledOn) != TCL_OK) {
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    if (ledOn) {
		kbControl.led_mode = LedModeOn;
	    } else {
		kbControl.led_mode = LedModeOff;
	    }

	    XChangeKeyboardControl(opts->display, KBLed|KBLedMode, &kbControl);
	}

	return TCL_OK;
    }

    if (strcmp(argv[2], "mapping") == 0) {
	if (argc > 4) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], " ?keymap?", (char *) NULL);
	    return TCL_ERROR;
	}

	/* Get the allowed keycode range. */

	XDisplayKeycodes(opts->display, &minKeycode, &maxKeycode);

	if (argc == 3) {
	    /* Get the full keysym list. */

	    keycodeCount = maxKeycode - minKeycode + 1;
	    keySymList = XGetKeyboardMapping(opts->display,
			minKeycode, keycodeCount, &keysymsPerKeycode);

	    /* Generate the result as keycode/keysymlist pairs. */

	    Tcl_DStringInit(&ds);
	    keysymIdx = 0;
	    for(i = 0; i < keycodeCount; i++) {
		sprintf(junkStr, "%1d", minKeycode + i);
		Tcl_DStringAppendElement(&ds, junkStr);
		Tcl_DStringStartSublist(&ds);
		for(j = 0; j < keysymsPerKeycode; j++) {
		    keysym = keySymList[keysymIdx];
		    Tcl_DStringAppendElement(&ds, XKeysymToString(keysym));
		    keysymIdx += 1;
		}
		Tcl_DStringEndSublist(&ds);
	    }

	    /* Clean up and return. */

	    XFree(keySymList);
	    Tcl_DStringResult(interp, &ds);
	    Tcl_DStringFree(&ds);
	    return TCL_OK;
	}

	/*
	 * Must be setting the keymap; split out the keycode/keysym
	 * map into a list, and do a prescan to get the map size.
	 */

	minMapKeycode = 255;
	maxMapKeycode = 8;
	maxKeysymsPerKeycode = 0;

	if (Tcl_SplitList(interp, argv[4], &nItems, &itemList) != TCL_OK) {
	    return TCL_ERROR;
	}

	if (nItems%2 != 0) {
	    ckfree((VOID *) itemList);
	    Tcl_SetResult(interp,
			"keyboard map must have an even number of elements",
								TCL_STATIC);
	    return TCL_ERROR;
	}

	for(i = 0; i < nItems; i += 2) {
	    if (Tcl_GetInt(interp, itemList[i], &keyCode) != TCL_OK) {
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    if (keyCode < minKeycode || keyCode > maxKeycode) {
		ckfree((VOID *) itemList);
		Tcl_AppendResult(interp, "keycode ",
				itemList[i], " is out of range", (char *) 0);
		return TCL_ERROR;
	    }

	    if (keyCode < minMapKeycode) {
		minMapKeycode = keyCode;
	    }

	    if (keyCode > maxMapKeycode) {
		maxMapKeycode = keyCode;
	    }

	    if (Tcl_SplitList(interp, itemList[i + 1],
				&nKeysyms, &keysymNameList) != TCL_OK) {
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    if (nKeysyms > maxKeysymsPerKeycode) {
		maxKeysymsPerKeycode = nKeysyms;
	    }

	    ckfree((VOID *) keysymNameList);
	}

	/* Get the current keyboard map. */

	currentKeycodeCount = maxKeycode - minKeycode + 1;
	currentSymList = XGetKeyboardMapping(opts->display,
			minKeycode, currentKeycodeCount,
			&currentKeysymsPerKeycode);
	if (maxKeysymsPerKeycode < currentKeysymsPerKeycode) {
	    maxKeysymsPerKeycode = currentKeysymsPerKeycode;
	}

	/* Build the keyboard map. Start with an empty map. */

	newMapNumEntries = maxMapKeycode - minMapKeycode + 1;
	newMapSize = newMapNumEntries*maxKeysymsPerKeycode;
	newKeyMap = (KeySym *) ckalloc(newMapSize*sizeof(KeySym));
	if (newKeyMap == (KeySym *) NULL) {
	    ckfree((VOID *) itemList);
	    XFree(currentSymList);
	    Tcl_SetResult(interp, "out of memory", TCL_STATIC);
	    return TCL_ERROR;
	}

	for (i = 0; i < newMapSize; i++) {
	    newKeyMap[i] = NoSymbol;
	}

	/* Put in the entries from the old map. */

	offset = minMapKeycode - minKeycode;
	for(i = 0; i < newMapNumEntries; i++) {
	    for(j = 0; j < currentKeysymsPerKeycode; j++) {
		newKeyMap[i + j] = currentSymList[i + offset + j];
	    }
	}
	XFree(currentSymList);

	/* Put in the entries defined in the mapping list. */

	for(i = 0; i < nItems; i += 2) {
	    if (Tcl_GetInt(interp, itemList[i], &keyCode) != TCL_OK) {
		ckfree((VOID *) newKeyMap);
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    if (Tcl_SplitList(interp, itemList[i + 1],
				&nKeysyms, &keysymNameList) != TCL_OK) {
		ckfree((VOID *) itemList);
		return TCL_ERROR;
	    }

	    for(j = 0; j < nKeysyms; j++) {
		newKeysym = XStringToKeysym(keysymNameList[j]);
		if (newKeysym == NoSymbol && keysymNameList[j][0] != '\0') {
		    Tcl_AppendResult(interp,
				"unknown keysym name \"", keysymNameList[j],
				"\"", (char *) 0);
		    ckfree((VOID *) keysymNameList);
		    ckfree((VOID *) itemList);
		    return TCL_ERROR;
		}
		keycodeOffset = keyCode - minMapKeycode;
		newKeyMap[keycodeOffset*maxKeysymsPerKeycode + j] = newKeysym;
	    }

	    ckfree((VOID *) keysymNameList);
	}
	ckfree((VOID *) itemList);

	/* Change the map. */

	XChangeKeyboardMapping(opts->display, minMapKeycode,
			maxKeysymsPerKeycode, newKeyMap, newMapNumEntries);

	/* Clean up and return. */

	ckfree((VOID *) newKeyMap);
	return TCL_OK;
    }

    if (strcmp(argv[2], "state") == 0) {
	if (argc != 3) {
	    Tcl_AppendResult(interp, "Usage: ", argv[0], " ", argv[1],
			" ", argv[2], (char *) NULL);
	    return TCL_ERROR;
	}

	XQueryKeymap(opts->display, keyStateReturn);
	/* Write the autorepeat map into interp. */

	for(keycode = 8; keycode <= 255; keycode++) {
	    idx = keycode/8;
	    mask = 1 << (keycode & 0x07);
	    if (keyStateReturn[idx] & mask) {
		sprintf(junkStr, "%1d", keycode);
		Tcl_AppendElement(interp, junkStr);
	    }
	}

	return TCL_OK;
    }

    Tcl_AppendResult(interp, "Usage: ", argv[0], " ",
		 argv[1], " ", argv[2], " set|get ...", (char *) NULL);
    return TCL_ERROR;
}

